# Scapy
import scapy.all as sc
from scapy.all import *

import socket

import threading
from queue import Queue

import dash # Dash framework
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input,Output

import pandas as pd # Pandas

import plotly.graph_objects as go # Plotly

# inicializacia Pandas DataFrame,index = netid
ip_data = pd.DataFrame(columns = ["netid","IP","MAC"])
port_data = pd.DataFrame(columns = ["netid","IP","PORT","SERVICE"])
ip_data.set_index("netid",inplace = True)
port_data.set_index("netid",inplace = True)

socket.setdefaulttimeout(2) # 2s socket timeout
m_lock = threading.Lock() # jednoduchy zamok

SYNACK = 0x12

PORTSRV = {
    7:  "ECHO protokol",
    13: "DAYTIME protokol",
    20: "FTP prenos dát",
    21: "FTP správa",
    22: "SSH protokol",
    23: "Telnet protokol",
    25: "SMTP protokol",
    37: "TIME protokol",
    42: "WINS služba",
    43: "WHOIS protokol",
    52: "XNS Time protokol",
    53: "DNS služba",
    56: "RAP protokol",
    67: "DHCP protokol",
    68: "DHCP protokol",
    69: "TFTP protokol",
    79: "FINGER protokol",
    80: "HTTP protokol",
    88: "Kerberos autentifikácia",
    107:"Vzdialená Telnet služba",
    109:"POP2 protokol",
    110:"POP3 protokol",
    123:"NTP protokol",
    137:"NetBIOS Name služba",
    138:"NetBIOS Datagram služba",
    139:"NetBIOS Session služba",
    143:"IMAP protokol",
    156:"SQL služba",
    161:"SNMP protokol",
    162:"SNMP protokol",
    170:"Network PostScript",
    194:"IRC služba",
    220:"IMAP v3",
    389:"LDAP protokol",
    443:"HTTPS protokol",
    465:"SMTP over TLS/SSL",
    520:"RIP protokol",
    546:"DHCPv6 klient",
    547:"DHCPv6 server",
    554:"RTSP protokol",
    631:"CUPS služba",
    636:"LDAP over TLS/SSL",
    901:"Samba WAT",
    903:"VMware konzola",
    989:"FTP data over TLS/SSL",
    990:"FTP control over TLS/SSL",
    995:"POP3 over TLS/SSL"
}

conf.verb = 0 # minimalizacia Scapy verbose info

def ARP_scan(dst_ip):
    ARP_req_frame = sc.ARP(pdst = dst_ip) # vytvorenie ARP Request frame
    #print(sc.ls(sc.ARP())) # polozky triedy ARP
    #print(arp_req_frame.summary()) # ARP Request frame
    #print(arp_req_frame.show()) # ARP Request frame

    broadcast_ether_frame = sc.Ether(dst = "ff:ff:ff:ff:ff:ff") # vytvorenie broadcast Ethernet frame
    #print(sc.ls(sc.Ether())) # polozky triedy Ether
    #print(broadcast_ether_frame.summary()) # broadcast Ethernet frame
    #print(broadcast_ether_frame.show()) # broadcast Ethernet frame
    
    broadcast_ether_ARP_req_frame = broadcast_ether_frame / ARP_req_frame # kombinacia broadcast Ethernet frame a ARP Request frame
    #print(broadcast_ether_ARP_req_frame.summary()) # broadcast Ethernet frame + ARP Request frame
    #for req in broadcast_ether_ARP_req_frame:
    #    req.show()

    #print("Skenujem IP:",d_ip)
    answ_list = sc.srp(broadcast_ether_ARP_req_frame,timeout = 2,verbose = False)[0] # zaslanie skombinovaneho framu
    if answ_list is not None:
        for i in range(0,len(answ_list)): # ak zariadenie odpovedalo
            ip_data.loc[answ_list[i][1].psrc] = (answ_list[i][1].psrc,answ_list[i][1].hwsrc)

def ICMP_scan(dst_ip,dst_port):
    try:
        pkt = sc.IP(dst = dst_ip)/sc.TCP(dport = dst_port,flags = "S")

        #print("Skenujem IP:",d_ip)
        answ_list = sc.sr(pkt,timeout = 2,verbose = False)[0]
        if answ_list is not None:
            for i in range(0,len(answ_list)): # ak zariadenie odpovedalo
                ip_data.loc[answ_list[i][1].psrc] = (answ_list[i][1].psrc,answ_list[i][1].hwsrc)
    except:
        pass

IPq = Queue()
PORTq = Queue()
IPscandone = True
PORTscandone = True

def IPscanner():
    d_IP = IPq.get()
    ARP_scan(d_IP)
    #ICMP_scan(d_IP,80)
    IPq.task_done()

def scan_IPs(f3o,start,end):
    global IPscandone

    if IPscandone == True:
        IPscandone = False
        print("IP sken začatý")

        ip_data.drop(ip_data.index,inplace = True) # vymazanie DataFrame

        for i in range(start,end + 1):
            d_ip = f3o + str(i)
            IPq.put(d_ip)

        for x in range(start,end + 1):
            t = threading.Thread(target = IPscanner)
            t.daemon = True
            t.start()

        IPq.join()
        IPscandone = True
        print("IP sken ukončený")

def socket_scan(dst_ip,dst_port):
    #print("Skenujem IP:PORT",dst_ip,dst_port)
    s = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
    try:
        con = s.connect((dst_ip,dst_port))
        with m_lock: # m_lock.acquire() - try ... - m_lock.release()
            #print("port",dst_port,"je otvorený")
            service = PORTSRV.get(dst_port,"Neznáma služba")
            netid = dst_ip + ":" + str(dst_port)
            port_data.loc[netid] = (dst_ip,dst_port,service)
        con.close()
    except:
        pass

def TCP_connect_scan(dst_ip,dst_port): # SYN -> SYNACK -> ACKRST
    src_port = sc.RandShort()
    try:
        #print("Skenujem IP:PORT",dst_ip,dst_port)
        resp = sc.sr1(sc.IP(dst = dst_ip)/sc.TCP(sport = src_port,dport = dst_port,flags = "S"),timeout = 2,verbose = False)
        if resp is not None:
            if resp.haslayer(sc.TCP):
                pktflags = resp.getlayer(sc.TCP).flags
                if pktflags == SYNACK:
                    #print("port",dst_port,"je otvorený")
                    resp = sc.sr(sc.IP(dst = dst_ip)/sc.TCP(sport = src_port,dport = dst_port,flags = "AR"),timeout = 2,verbose = False)
                    service = PORTSRV.get(dst_port,"Neznáma služba")
                    netid = dst_ip + ":" + str(dst_port)
                    port_data.loc[netid] = (dst_ip,dst_port,service)
            elif resp.haslayer(sc.ICMP):
                if (int(resp.getlayer(sc.ICMP).type) == 3) and int(resp.getlayer(sc.ICMP).code) in [1,2,3,9,10,13]:
                    print("port",dst_port,"je filtrovaný")
    except:
        pass

def TCP_stealth_scan(dst_ip,dst_port): # SYN -> SYNACK -> RST
    src_port = sc.RandShort()
    try:
        #print("Skenujem IP:PORT",dst_ip,dst_port)
        resp = sc.sr1(sc.IP(dst = dst_ip)/sc.TCP(sport = src_port,dport = dst_port,flags = "S"),timeout = 2,verbose = False)
        if resp is not None:
            if resp.haslayer(sc.TCP):
                pktflags = resp.getlayer(sc.TCP).flags
                if pktflags == SYNACK:
                    #print("port",dst_port,"je otvorený")
                    resp = sc.sr(sc.IP(dst = dst_ip)/sc.TCP(sport = src_port,dport = dst_port,flags = "R"),timeout = 2,verbose = False)
                    service = PORTSRV.get(dst_port,"Neznáma služba")
                    netid = dst_ip + ":" + str(dst_port)
                    port_data.loc[netid] = (dst_ip,dst_port,service)
            elif resp.haslayer(sc.ICMP):
                if (int(resp.getlayer(sc.ICMP).type) == 3) and int(resp.getlayer(sc.ICMP).code) in [1,2,3,9,10,13]:
                    print("port",dst_port,"je filtrovaný")
    except:
        pass

def PORTscanner(d_IP):
    d_PORT = PORTq.get()
    #socket_scan(d_IP,d_PORT)
    TCP_connect_scan(d_IP,d_PORT)
    #TCP_stealth_scan(d_IP,d_PORT)
    PORTq.task_done()

def scan_PORTs(start,end):
    global PORTscandone

    if PORTscandone == True:
        PORTscandone = False
        print("PORT sken začatý")

        port_data.drop(port_data.index,inplace = True) # vymazanie DataFrame

        for d_ip in ip_data.IP:

            for d_port in range(start,end + 1):
                PORTq.put(d_port)

            for x in range(start,end + 1):
                t = threading.Thread(target = PORTscanner,args = (d_ip,))
                t.daemon = True
                t.start()

            PORTq.join()
        
        PORTscandone = True
        print("PORT sken ukončený")

mIP = sc.get_if_addr(conf.iface) # moja IP adresa
s = mIP.split(".")
f3o = s[0] + "." + s[1] + "." + s[2] + "." # prve 3 oktety IP adresy

# inicializacia Dash aplikacie
app = dash.Dash(__name__)

# definicia Dash layoutu
app.layout = html.Div(
    # 2 riadky
    children = [html.Div(className = 'row', # 1. riadok ----------------------------------------------------------------------
                         children = [# 1. stlpec
                                     html.Div(children = [# 1. riadok
                                                          html.H1('Skener dátovej siete'),
                                                          html.Div(className = 'block_container',
                                                                   style = {'margin' : 20},
                                                                   children = ['Moja IP adresa je:',
                                                                               html.Div(className = 'IPs',children = [ mIP ])]
                                                          ),
                                                          html.Div(className = 'block_container',
                                                                   style = { 'margin' : 20 },
                                                                   children = [
                                                                       'Skenujem IP adresy: od',
                                                                       html.Div(className = 'IPs',children = [ f3o ]),
                                                                       dcc.Input(style = {'width' : 70,'margin-right' : 5},
                                                                                 id = 'startIP',type = 'number',value = '1',
                                                                                 min = '1',max = '254',debounce = True),
                                                                       'do',
                                                                       html.Div(className = 'IPs',children = [ f3o ]),
                                                                       dcc.Input(style = {'width' : 70,'margin-right' : 5},
                                                                                 id = 'endIP',type = 'number',value = '254',
                                                                                 min = '1',max = '254',debounce = True),
                                                                       html.Div(id = 'priebeh1'),
                                                                       dcc.Interval(
                                                                          id = 'tinterval1',
                                                                          interval = 1000, # updatuje n_intervals kazdu 1s
                                                                          n_intervals = 0
                                                                       )
                                                                   ]
                                                          ),
                                                          html.Div(className = 'block_container',
                                                                   style = {'margin' : 20},
                                                                   children = [
                                                                       'Skenujem PORTY: od',
                                                                       dcc.Input(style = {'width' : 70,
                                                                                          'margin-left' : 5,
                                                                                          'margin-right' : 5},
                                                                                 id = 'startPORT',type = 'number',value = '1',
                                                                                 min = '1',max = '65535',debounce = True),
                                                                       'do',
                                                                       dcc.Input(style = {'width' : 70,
                                                                                          'margin-left' : 5,
                                                                                          'margin-right' : 5},
                                                                                 id = 'endPORT',type = 'number',value = '500',
                                                                                 min = '1',max = '65535',debounce = True),
                                                                       html.Div(id = 'priebeh2')
                                                                   ]
                                                          ),
                                                          html.Div(style = {'margin' : 20},
                                                                   children = [
                                                                       html.Button('Skenuj',id = 'Skenuj',n_clicks = 0,
                                                                                   style = { 'margin-right' : 5 })
                                                                   ]
                                                          )
                                              ]
                                     )
                         ]
                ),
                html.Div(className = 'row', # 2. riadok ----------------------------------------------------------------------
                         children = [# 1. stlpec
                                     html.Div(className = 'four columns',style = {'margin-left' : 20},
                                              children = [# 1. riadok
                                                          dcc.Graph(id = 'table1',
                                                                    figure = go.Figure(data = None,
                                                                                       layout = dict(xaxis = {'visible' : False},
                                                                                                     yaxis = {'visible' : False},
                                                                                                     template = 'plotly_dark',
                                                                                                     paper_bgcolor = 'rgba(0,0,0,0)',
                                                                                                     plot_bgcolor = 'rgba(0,0,0,0)')
                                                                    )
                                                          )
                                              ]
                                     ),
                                     # 2. stlpec
                                     html.Div(className = 'five columns',style = {'margin-left' : 20},
                                              children = [# 1. riadok
                                                          dcc.Graph(id = 'table2',
                                                                    figure = go.Figure(data = None,
                                                                                       layout = dict(xaxis = {'visible' : False},
                                                                                                     yaxis = {'visible' : False},
                                                                                                     template = 'plotly_dark',
                                                                                                     paper_bgcolor = 'rgba(0,0,0,0)',
                                                                                                     plot_bgcolor = 'rgba(0,0,0,0)')
                                                                    )
                                                          ),
                                                          dcc.Interval(
                                                              id = 'tinterval2',
                                                              interval = 5000, # updatuje n_intervals kazdych 5s
                                                              n_intervals = 0
                                                          )
                                              ]
                                     )
                         ]
                )
    ]
)
    
# Callback funkcia pre tabulku 1
@app.callback(Output('table1','figure'),[Input('Skenuj','n_clicks'),Input('startIP','value'),Input('endIP','value'),
                                         Input('startPORT','value'),Input('endPORT','value')])
def update_table1(n,sIP,eIP,sP,eP):
    try:
        startIP = int(sIP)
        endIP = int(eIP)
        if endIP >= startIP:
            scan_IPs(f3o,startIP,endIP) # IP sken
            if IPscandone == True: # ak bol ukonceny IP sken
                startPORT = int(sP)
                endPORT = int(eP)
                if endPORT >= startPORT:
                    scan_PORTs(startPORT,endPORT) # PORT sken
    except:
        pass

    # tabulka
    fig = go.Figure(data=[go.Table(header = dict(values = ['IP','MAC']),
                                   cells = dict(values = [ip_data.IP,ip_data.MAC],height = 30),
                                   columnwidth = [100,100])
                         ]
                   )
    # nastavenie stylu tabulky
    fig.update_layout(template = 'plotly_dark',
                      font_size = 15,
                      paper_bgcolor = 'rgba(0,0,0,100)',
                      margin = {'l' : 10,'r' : 10,'t' : 10,'b' : 10},
                      showlegend = False)
    # aktualizacia dcc.Graph(figure)
    return fig

# Callback funkcia pre tabulku 2
@app.callback(Output('table2','figure'),[Input('tinterval2','n_intervals')])
def update_table2(n):
    # tabulka
    fig = go.Figure(data=[go.Table(header = dict(values = ['IP','PORT','SLUŽBA / PROTOKOL']),
                                   cells = dict(values = [port_data.IP,port_data.PORT,port_data.SERVICE],height = 30),
                                   columnwidth = [100,100,180])
                         ]
                   )
    # nastavenie stylu tabulky
    fig.update_layout(template = 'plotly_dark',
                      font_size = 15,
                      paper_bgcolor = 'rgba(0,0,0,100)',
                      margin = {'l' : 10,'r' : 10,'t' : 10,'b' : 10},
                      showlegend = False)
    # aktualizacia dcc.Graph(figure)
    return fig

@app.callback([Output('priebeh1','children'),Output('priebeh2','children')],[Input('tinterval1','n_intervals')])
def update_priebeh1a2(n):
    t1 = ' IP sken prebieha...'
    t2 = ' PORT sken prebieha...'

    if IPscandone == True:
        t1 = ' IP sken ukončený'
    if PORTscandone == True:
        t2 = ' PORT sken ukončený'
        
    return [t1,t2]

if __name__ == "__main__":
    # spustenie Dash aplikacie
    app.run_server(debug = True)